"""
Launch Browsers if they exist, to initialize
the preferences so that the CA can be installed.
"""
import time
from pathlib import Path


def configure_browsers(ctx, name):
    """
    Configure all browsers on the Linux platform
    """
    brave(ctx)
    chrome(ctx)
    firefox(ctx)
    cert_db(ctx, name)


def brave(ctx):
    """
    Check if Brave is installed and initialized.
    """
    check_installed(
        ctx,
        'brave-browser',
        install_path='/opt/brave.com',
        prefs_path='~/.config/BraveSoftware',
    )


def chrome(ctx):
    """
    Check if Google Chrome is installed and initialized.
    """
    check_installed(
        ctx,
        'google-chrome',
        install_path='/opt/google',
        prefs_path='~/.config/google-chrome',
    )


def firefox(ctx):
    """
    Check if Firefox is installed and initialized.
    """
    check_installed(ctx, 'firefox', install_path='/lib/firefox', prefs_path='~/.mozilla/firefox')


def cert_db(ctx, name):
    """
    Install CA in cert9.db
    """
    db_files = Path('~/').expanduser().rglob('cert9.db')
    cert_path = ctx['cert_path']

    for db in db_files:
        print(f'\nAdding CA to {db}')
        cert_dir = db.parent
        ctx.run(f'certutil -A -n "{name}" -t "TCu,Cu,Tu" -i {cert_path} -d sql:{cert_dir}')


def check_installed(ctx, name, install_path, prefs_path):
    """
    Check if browser is installed and preferences initialized.
    """
    if not Path(install_path).is_dir():
        return False

    if not Path(prefs_path).expanduser().is_dir():
        ctx.run(f'{name} &')
        time.sleep(5)

        # when launching brave-browser or google-chrome, only kill prefixed part
        name = name.split('-')[0]
        ctx.run(f'pkill -9 -f {name}')

    return True


__all__ = (
    'configure_browsers',
    'brave',
    'chrome',
    'firefox',
    'cert_db',
)
