from cryptography.fernet import InvalidToken
from lcconfig.encryption import decrypt, encrypt
import os
import re

from aspenctl import settings

try:
    import argon2
    from keyrings.cryptfile.cryptfile import CryptFileKeyring as EncryptedFileBase
except (ImportError, ModuleNotFoundError):
    from keyrings.cryptfile.file import EncryptedKeyring as EncryptedFileBase


class EncryptedFileBackend(EncryptedFileBase):
    """
    Use encrypted file for environments where credential managers
    aren't available, such as WSL and cygwin.

    Overriding to save password to environment variable, so
    that the user doesn't need to enter the password for keyring
    read and writes in the same set of tasks runs.
    """

    def _get_new_password(self):
        """
        Add some password strength checking, and save password to environment.
        """
        # Make sure that users enter secure password
        password = None
        while not password:
            pwd = super()._get_new_password()
            if len(pwd) < 10:
                print('Password length must be at least 10 characters')
                continue

            if not re.match(r'((?=.*\d)(?=.*[a-z])(?=.*[A-Z]).{6,20})', pwd):
                print('Passwords must contain at lease 1 digit and upper-case character')
                continue

            password = pwd

        self.add_to_environment(password)
        return password

    def _unlock(self):
        """
        Try to unlock keyring with environment variable value before prompting user
        for password entry.
        """
        try:
            self._keyring_key = self.load_from_environment()
            return super()._unlock()
        except ValueError:
            pass

        self.add_to_environment('')
        super()._unlock()
        self.add_to_environment(password=self._keyring_key)

    def add_to_environment(self, password: str):
        """
        Save to environment variable to minimize times we need to prompt the user.
        """
        # If not password, ensure the value is deleted, so that next time
        # we prompt the user for a new password
        if not password:
            return os.environ.pop(settings.ASPEN_SESSION, None)

        try:
            os.environ[settings.ASPEN_SESSION] = encrypt(password)
        except Exception:
            pass

    def load_from_environment(self):
        """
        Retrieve keyring key from environment, or reset environment if
        unable to decrypt current value.
        """
        try:
            return decrypt(os.environ.get(settings.ASPEN_SESSION, ''))
        except (InvalidToken, TypeError):
            self.add_to_environment('')
