from cryptography.fernet import InvalidToken
from datetime import datetime, timedelta
from lcconfig import decrypt, encrypt
import jwt
from .keyring_service import keyring_svc
from keyring.errors import (
    KeyringError,
    NoKeyringError,
    KeyringLocked,
    PasswordDeleteError,
)
from typing import NamedTuple
from aspenctl import settings


class Secret(NamedTuple):
    name: str
    valid: bool


def load_token(token_name: str) -> Secret:
    """
    Load token from keyring and check if token has not expired.
    """
    token = load_secret(token_name)
    if not token:
        return Secret('', False)

    try:
        token_data = jwt.decode(token, options={"verify_signature": False})
        token_expiration = datetime.fromtimestamp(token_data['exp'])
        return Secret(token, token_expiration >= datetime.now() + timedelta(minutes=1))
    except jwt.exceptions.PyJWTError as e:
        print(e)
        return Secret('', False)


def load_secret(name: str) -> str:
    """
    Get secret from keyring.
    """
    if not name or settings.IS_DRAWBRIDGE_OR_SERVER:
        return ''

    try:
        secret = keyring_svc.get_password('system', name)
    except NoKeyringError:
        raise SystemExit('Please install keyring package for your OS.')
    except KeyringLocked:
        raise SystemExit('Unable to unlock your OS keyring.')
    except KeyringError as e:
        raise SystemExit(f'A general keyring error occurred: {e}')
    except ValueError as e:
        if 'mac check failed' in str(e).lower():
            raise SystemExit('Incorrect keyring password')
        raise

    try:
        return decrypt(secret)
    except (InvalidToken, TypeError):
        pass

    return ''


def save_secret(name: str, value: str):
    """
    Save secret to keyring.
    """
    if settings.IS_DRAWBRIDGE_OR_SERVER:
        return

    try:
        keyring_svc.set_password('system', name, encrypt(value))
    except KeyringLocked:
        raise SystemExit(f'Unable to save {name} secret to your System Keyring!')


def purge_keyring_secrets(ctx):  # noqa
    """
    Purge all AspenCtl secrets from keyring.
    """
    for key in (
            settings.ASPEN_USER_PK_PWD,
            settings.ASPEN_ACCESS_TOKEN,
            settings.ASPEN_REFRESH_TOKEN,
    ):
        try:
            keyring_svc.delete_password('system', key)
        # servers may not have keyring
        except NoKeyringError:
            return
        except PasswordDeleteError:
            pass


__all__ = (
    'load_token',
    'load_secret',
    'save_secret',
    'purge_keyring_secrets',
)
