from getpass import getpass
from invoke import Context
import jwt
import os
from pylogcabin import AspenAPI as AspenAPIBase
from pylogcabin.exceptions import (
    LCHTTPError,
    ConsoleAPIError,
    ConsoleAPIAuthenticationError,
    ConsoleAPIAuthorizationError,
)
from .configuration import AspenCtlSettingsConfig, ConsoleSettingsConfig
from .encryption import load_token, save_secret
from .typehints import ApiUser, AspenAccountConfig, Tokens
from . import settings


class AspenAPI(AspenAPIBase):
    """
    Extend to add handy registration method.
    """

    def get_initial_token(self, username: str, password: str) -> Tokens:
        """
        Get JWT authenticating response from server, so the token
        can be saved to the configuration file for a period of time
        to prevent the user from continually entering credentials.
        """
        credentials = {'username': username, 'password': password}
        return self._token_request(url_path='/latchstring/jwt/token/', post_data=credentials)

    def get_refreshed_token(self, refresh_token: str) -> Tokens:
        """
        Get JWT authenticating response from server based on refresh_token.
        """
        credentials = {'refresh': refresh_token}
        url = '/latchstring/jwt/token/refresh/'
        return self._token_request(url_path=url, post_data=credentials)

    def _token_request(self, url_path: str, post_data: dict = None) -> Tokens:
        self.headers.pop('Authorization', False)
        self.url = f'{self.SERVER}{url_path}'
        self.headers['Content-Type'] = 'application/json'

        try:
            return self.post(post_data=post_data).json()
        except ConsoleAPIAuthenticationError:
            raise SystemExit('Login failed! Check your username / password credentials.')
        except ConsoleAPIAuthorizationError:
            raise SystemExit('Authorization denied with the user credentials provided.')
        except ConsoleAPIError as e:
            print(e)
            raise SystemExit('An error occurred; unable to contact Aspen Server.')

    def http_request(self, *args, **kwargs):
        """
        Override here to catch hari kiri status code, which is sent from Aspen Server
        if this device has been deactivated.
        """
        try:
            return super().http_request(*args, **kwargs)
        except LCHTTPError as e:
            if e.response.status != settings.POISON_PILL_STATUS_CODE:
                raise

            from aspenctl.tasks.register import unregister
            ctx = Context()
            unregister(ctx)
            raise SystemExit('Device has been de-registered from Aspen')


def get_api(ctx, account_cid='') -> ApiUser:
    """
    Return user and Aspen API class.
    """
    # Get initial token, saving initial config file if necessary
    access_token = aspen_api_token(ctx)

    try:

        if settings.IS_DRAWBRIDGE_OR_SERVER:
            cfg = ConsoleSettingsConfig()
            data = cfg.as_typed_dict(settings.SYSTEM_SECTION)
            account_cid = account_cid or data.get('cid', '')

        if not account_cid:
            if settings.IS_DRAWBRIDGE_FILTER:
                raise SystemExit('Complete Drawbridge installation before configuring Aspen.')

            cfg = AspenCtlSettingsConfig()
            data = cfg.as_typed_dict(settings.ACCOUNT_SECTION)
            account_cid = account_cid or data.get('cid', '')

        if not account_cid:
            raise SystemExit(
                'User Account creation not complete. Consider running "aspen register.unregister"'
            )

        return ApiUser(AspenAPI(token=access_token), account_cid)

    except Exception:
        raise SystemExit(f'Unable to initialize API class')


def aspen_api_token(ctx) -> str:
    """
    Load credentials for API use, in one of the following ways:
       - Load access token from config file
       - Refresh token and save to config file
       - Prompt user for credentials to retrieve tokens
    """
    if settings.IS_DRAWBRIDGE_OR_SERVER:
        # Try DrawBridge config file first, so that we can run the aspenctl utility
        # on DrawBridge systems via the system token.
        # TODO consider using keyring with a plain-text backend, since values are encoded anyway?
        console_cfg = ConsoleSettingsConfig()
        console_data = console_cfg.as_typed_dict(settings.SYSTEM_SECTION)

        # If no auth_token found, allow the device-specific authentication procedure to complete.
        # The server registration process will save the auth_token to config file, for future use.
        if 'auth_token' in console_data:
            return console_data['auth_token']

    token = load_token(settings.ASPEN_ACCESS_TOKEN)
    if token.valid:
        return token.name

    return get_renewable_access_tokens(ctx)


def get_renewable_access_tokens(ctx) -> str:
    """
    Get new token and save to disk.

    If refresh_token is invalid or unset, prompt for username / password.
    """
    new_token_data = {}

    refresh_token = load_token(settings.ASPEN_REFRESH_TOKEN)
    if refresh_token.valid:
        try:
            new_token_data = AspenAPI().get_refreshed_token(refresh_token.name)
        except ConsoleAPIError:
            pass

    if not new_token_data:
        # Save creds to environment so that when registering Servers the user
        # doesn't have to authenticate twice to perform the registration / saving
        username = os.getenv('username', '')
        password = os.getenv('password', '')

        if not username or not password:
            print('Enter your Clavis SSO username and password to access Aspen\n')
        while not username:
            username = os.environ['username'] = input('Username: ')
        while not password:
            password = os.environ['password'] = getpass('Password: ')

        try:
            new_token_data = AspenAPI().get_initial_token(username, password)
        except ConsoleAPIError:
            pass

    if not new_token_data:
        return ''

    # Extract user's CID from token claims
    try:
        token_data = jwt.decode(new_token_data['access'], options={"verify_signature": False})
    except jwt.exceptions.DecodeError:
        print('Unable to decode token data')
        return ''

    persist_token = ctx.get(settings.PERSIST_TOKEN, True)
    if persist_token:
        new_token_data['cid'] = token_data['cid']
        save_aspenctl_config(data=AspenAccountConfig(**new_token_data))

    return new_token_data['access']


def save_aspenctl_config(data: AspenAccountConfig):
    """
    Save AspenCtl Account data to configuration file.
    Expecting dictionary of token data from Aspen server.
    """
    save_secret(settings.ASPEN_ACCESS_TOKEN, data['access'])

    try:
        # If the response includes a new refresh token, save it.
        # Server decides whether to refresh the tokens or not.
        save_secret(settings.ASPEN_REFRESH_TOKEN, data['refresh'])
    except KeyError:
        pass

    cfg = AspenCtlSettingsConfig()

    config_data = {
        'cid': data['cid'],
    }

    cfg.save_section(settings.ACCOUNT_SECTION, config_data)

    return config_data


__all__ = (
    'AspenAPI',
    'get_api',
)
