from cryptography.hazmat.primitives.serialization import Encoding
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes
import csv
from ipaddress import IPv4Address
import os
from pathlib import Path
from pylogcabin.exceptions import (
    LCHTTPError,
    ConsoleAPIAuthenticationError,
    ConsoleAPINotFoundError,
)

from aspenctl.configuration import ConsoleSettingsConfig
from aspenctl import settings
from aspenctl.network import get_api
from aspenctl.validation import display_errors
from .keys import generate_key


def generate_csr(
    common_name: str,
    country: str,
    state: str,
    locality: str,
    organization: str,
    email: str,
    dnsnames=(),
    ips=(),
):
    """
    Generate Certificate Signing request
    """
    altnames = []
    for dns in dnsnames:
        altnames.append(x509.DNSName(dns))

    for ip in ips:
        altnames.append(x509.IPAddress(IPv4Address(ip)))

    subject_name = x509.Name([
        # Include all the details to be shown in the cert
        x509.NameAttribute(NameOID.COUNTRY_NAME, country.upper()),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state.upper()),
        x509.NameAttribute(NameOID.LOCALITY_NAME, locality),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization),
        x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        x509.NameAttribute(NameOID.EMAIL_ADDRESS, email),
    ])

    csr = x509.CertificateSigningRequestBuilder(
        subject_name=subject_name,
    ).add_extension(
        x509.KeyUsage(
            digital_signature=True,
            content_commitment=False,
            key_encipherment=True,
            data_encipherment=True,
            key_agreement=True,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False,
        ),
        critical=True,
    ).add_extension(
        x509.BasicConstraints(ca=False, path_length=None), critical=True
    ).add_extension(
        x509.ExtendedKeyUsage([x509.OID_SERVER_AUTH, x509.OID_CLIENT_AUTH]), critical=True
    ).add_extension(
        x509.SubjectAlternativeName(altnames),
        critical=False,
    ).sign(
        generate_key(common_name),
        hashes.SHA512(),
    )

    save_cert_file(f'{common_name}.csr', csr.public_bytes(Encoding.PEM).decode('utf8'))

    return csr


# ---------------------------------------------------------------------------
def sign_csr(ctx, common_name: str, expires: int, ca_pk: int):
    """
    Sign a CSR and save certificate.
    """
    try:
        with open(f'{settings.ASPEN_CONFIG_DIR}/{common_name}.csr', 'r') as csrp:
            csr = csrp.read()
    except FileNotFoundError:
        raise SystemExit(f'No CSR found for {common_name}')
    except Exception as e:
        raise SystemExit(f'Failed to read CSR - {e}')

    api, _ = get_api(ctx)

    api.load('aspentls', 'certificateauthority')

    cfg = ConsoleSettingsConfig()
    data = cfg.as_typed_dict(settings.SYSTEM_SECTION)

    data = {
        'name': common_name,
        'pem': csr,
        'appliance': data.get('cid'),
        'expires': int(expires),
        'cert_type': 'webserver',
    }

    try:
        resp = api.post(ca_pk, subpath='sign_csr', post_data=data)
    except ConsoleAPIAuthenticationError:
        raise SystemExit('Invalid credentials; unable to sign CSR.')
    except ConsoleAPINotFoundError:
        raise SystemExit('No CA with this ID found on Aspen CA Server')
    except LCHTTPError as e:
        display_errors(e.response)
        raise SystemExit(f'An error occurred on the Aspen server: {e}')
    except Exception as e:
        raise SystemExit(f'Failed to sign CSR on Aspen CA Server: {e}')

    save_cert_file(f'{common_name}.pem', resp.json()['pem'])


# ---------------------------------------------------------------------------
def save_cert_file(name: str, data: str):
    """
    Save text data to file.
    """
    try:
        os.makedirs(f'{settings.ASPEN_CONFIG_DIR}', exist_ok=True)
        with open(f'{settings.ASPEN_CONFIG_DIR}/{name}', 'w') as cf:
            cf.write(data)
    except Exception as e:
        raise SystemExit(f'Unable to save {name} to disk: {e}')


# ---------------------------------------------------------------------------
def save_csr_conf(name: str, expires: int, ca_pk: int, location: str):
    """
    Save Certificate Signing Request common name and CA PK
    to a csv file so it can be read later and the same CSR
    automatically re-signed and renewed.
    """
    records = csr_history()
    new_record = {'name': name, 'expires': expires, 'ca': ca_pk, 'location': location}

    for i, record in enumerate(records):
        if record['name'] == name:
            records[i] = new_record
            break
    else:
        records.append(new_record)

    with open(f'{settings.ASPEN_CONFIG_DIR}/csr_history.csv', 'w') as ch:
        writer = csv.DictWriter(ch, fieldnames=list(new_record.keys()))
        writer.writeheader()

        for record in records:
            if Path(f'{record["location"]}/{name}.csr').exists():
                writer.writerow(record)


# ---------------------------------------------------------------------------
def csr_history():
    """
    Get list of Certificate Signing Requests on this machine.
    """
    rows = []

    try:
        with open(f'{settings.ASPEN_CONFIG_DIR}/csr_history.csv', 'r') as ch:
            reader = csv.DictReader(ch)
            for row in reader:
                row['expires'] = int(row['expires'])
                rows.append(row)
    except FileNotFoundError:
        print('No Certificate Signing Request history found')
        pass

    return rows


__all__ = (
    'generate_csr',
    'sign_csr',
    'save_csr_conf',
    'csr_history',
)
