from lcconfig import ConsoleSettingsConfig, strtobool
from aspenctl.validation import is_valid_email_address, is_valid_hostname, is_valid_ip_address
from aspenctl.ssh import get_ip_addresses as get_adapter_ip_addresses


def get_common_name():
    """
    Get Certificate Common Name from user.
    """
    common_name = ''

    try:
        while not common_name:
            cn = input('Common Name: ')
            if not is_valid_hostname(cn):
                print('\nInvalid Common Name. Try again.\n')
            else:
                common_name = cn
    except (EOFError, KeyboardInterrupt):
        print()
        raise SystemExit()

    return common_name


def get_alt_dns_names():
    """
    Get Alternate DNS names for this certificate.
    """
    dns_names = []

    more = add_more('Add alternate DNS name to cert?')

    while more:
        try:
            dns = input('DNS Name: ')
            if not is_valid_hostname(dns):
                print('\nInvalid DNS Name. Try again.\n')
            else:
                dns_names.append(dns)
        except (EOFError, KeyboardInterrupt):
            return dns_names

        more = add_more('\nAdd another alternate DNS name?')

    return dns_names


def get_ip_addresses():
    """
    Get IPv4 Addresses for this certificate.
    """
    ip_addresses = []

    adapter_ips = get_adapter_ip_addresses()

    if adapter_ips:
        print(f'\nSystem has {len(adapter_ips)} on network interfaces\n')
        for ip in adapter_ips:
            if add_more(f'Add {ip}?'):
                ip_addresses.append(ip)

    while add_more('\nAdd another IPv4 Address?'):
        try:
            ip = input('IPv4 Address: ')
            if not is_valid_ip_address(ip):
                print('\nInvalid IPv4 Address. Try again.\n')
            else:
                ip_addresses.append(ip)
        except (EOFError, KeyboardInterrupt):
            return ip_addresses

    return ip_addresses


def get_email(default=''):
    """
    Get Email from user.
    """
    email = ''
    msg = f'Support Email ({default}): ' if default else f'Support Email: '

    try:
        while not email:
            e = input(msg) or default
            if not is_valid_email_address(e):
                print('\nInvalid email address. Try again.\n')
            else:
                email = e
    except (EOFError, KeyboardInterrupt):
        print()
        raise SystemExit()

    return email


def get_value_from_user(name, default=''):
    """
    Get generic value from user.
    """
    value = ''

    msg = f'{name} ({default}): ' if default else f'{name}: '

    try:
        while not value:
            value = input(msg).strip() or default
    except (EOFError, KeyboardInterrupt):
        print()
        raise SystemExit()

    return value


def add_more(msg):
    """
    Prompt user to see if they want to add another value.
    """
    try:
        return strtobool(input(f'{msg} [y/n]: ').strip())
    except ValueError:
        print('\nInvalid response. Please enter "y" or "n"\n')
        return add_more(msg)
    except (EOFError, KeyboardInterrupt):
        print()
        raise SystemExit()


def set_ca_defaults():
    """
    Set config file defaults to auto-fill CSR requests.
    """
    data = {
        'organization': get_value_from_user('Organization / Company'),
        'locality': get_value_from_user('Locality'),
        'state': get_value_from_user('State'),
        'email': get_value_from_user('Support Email'),
    }

    cfg = ConsoleSettingsConfig()
    cfg.save_section('CA', data)


__all__ = (
    'get_common_name',
    'get_alt_dns_names',
    'get_email',
    'get_ip_addresses',
    'get_value_from_user',
    'set_ca_defaults',
)
