from aspenctl.configuration import AspenCtlSettingsConfig, ConsoleSettingsConfig
from system_env.environment import setup_portal_environment
import ifaddr
from pathlib import Path
import socket
from typing import List, Set, Union
from aspenctl import settings


def get_network_ssh_principals() -> List[str]:
    """
    Get DNS & IP addresses to include as
    SSH Principals in the signed public key.
    """
    principals = get_nginx_server_names()
    principals.update(get_ip_addresses())
    principals.update(get_config_file_principals())
    principals.add(socket.gethostname())

    return [principal for principal in principals if principal]


def get_config_file_principals() -> Set[str]:
    """
    Load principals from Device and Account config files.
    """
    principals = set()

    cfg = ConsoleSettingsConfig()
    cfg_data = cfg.as_typed_dict(settings.SYSTEM_SECTION)

    for key in ('adminurl', 'cid', 'name'):
        value = cfg_data.get(key) or ''
        if value:
            principals.add(value)

    extra_principals = str(cfg_data.get('principals') or '').split(',')
    principals.update(principal.strip() for p in extra_principals if (principal := p.strip()))

    if not settings.IS_DRAWBRIDGE_OR_SERVER:
        # Add user-specific CID, so that this device host identity keys
        # are tied to their owner, and can restrict access by that principal
        cfg = AspenCtlSettingsConfig()
        cfg_data = cfg.as_typed_dict(settings.ACCOUNT_SECTION)
        cid = cfg_data.get('cid')
        if cid:
            principals.add(cid)

    return principals


def get_ip_addresses() -> Set[str]:
    """
    Return routable IPv4 addresses from
    network adapters.
    """
    ips = set()

    for adapter in ifaddr.get_adapters():
        for ip in adapter.ips:
            if ip.network_prefix == 24:
                ips.add(ip.ip)

    return ips


def get_nginx_server_names() -> Set[str]:
    """
    Extract server names from nginx config files
    to add them as SSH Principals so that the DNS
    names can be used correctly for SSHing in.
    """
    server_names = set()

    portal_env = setup_portal_environment()

    for directory in (
            '/etc/nginx/sites-available',
            '/etc/nginx/conf.d',
            portal_env['CONF_DIR'],
            '/etc/compass',  # if location above is ever relocated here
    ):
        for config_file in Path(directory).glob('*.conf'):
            dns_names = nginx_server_names(config_file)
            server_names.update(dns_names)

    for local in ('localhost', '127.0.0.1'):
        try:
            server_names.remove(local)
        except KeyError:
            pass

    return server_names


def nginx_server_names(path: Union[str, Path]) -> Set[str]:
    """
    Parse single config file, extracting all dns names from server_name params

    server_name thinkwell.compassfoundation.systems draw.bridge thinkwell.myvision.id log.cabin;
    server_name passageway.myvision.id;
    """
    server_names = set()

    try:
        with open(path, 'r') as sc:
            for config_line in sc:
                line = config_line.strip()
                if line.startswith('server_name'):
                    dns_names = line[12:].strip(';')
                    server_names.update(dns.strip() for dns in dns_names.split())
    except FileNotFoundError:
        pass
    except PermissionError:
        print(f'PermissionError! Unable to open {path}')

    return server_names


__all__ = (
    'get_network_ssh_principals',
    'get_config_file_principals',
    'get_ip_addresses',
    'get_nginx_server_names',
    'nginx_server_names',
)
