from aspen_ssh.certificates import SSHCertificateType
from aspen_ssh.parser.exceptions import SSHCertificateParserError
from aspen_ssh.parser import SSHCertificate
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PrivateFormat,
    PublicFormat,
    BestAvailableEncryption,
    NoEncryption,
)
from datetime import datetime, timedelta
import getpass
from lcconfig import AspenCtlSettingsConfig
import os
from pathlib import Path
from pylogcabin import AspenAPI
from pylogcabin.exceptions import ConsoleAPINotFoundError, LCHTTPError
from invoke import Responder
import secrets
import stat
import tempfile
from typing import Optional, Set, Tuple, Union

from ..encryption import load_secret, save_secret
from ..network import get_api
from .principals import get_network_ssh_principals
from ..typehints import Key, OpenSshSigningRequestData
from ..validation import display_errors
from .. import settings


def refresh_host_key(ctx, cid: str, api: AspenAPI):
    """
    Create host key files for identifying hosts with the CA.
    """
    keys = generate_openssh_keys(encrypted=False)

    data = OpenSshSigningRequestData(
        public_key=keys.public.decode('utf8'),
        principals=get_network_ssh_principals(),
        cert_type=SSHCertificateType.HOST,
        duration=settings.HOST_IDENTITY_KEY_DURATION,
    )

    try:
        resp = api.post(cid, post_data=data, subpath=settings.REFRESH_SSH_KEY_PATH)
    except ConsoleAPINotFoundError:
        url = api.get_url(cid, subpath=settings.REFRESH_SSH_KEY_PATH)
        raise SystemExit(f'Unable to refresh host identity key URL not found {url!r}')
    except LCHTTPError as e:
        display_errors(e.response)
        raise SystemExit(f'Unable to refresh host identity key - Error: {e}')

    signed_key = resp.json()['signed_key'].encode('utf8')

    try:
        persist_host_keys(ctx, keys, signed_key=signed_key)
        ctx.run(settings.SSH_SERVICE_CMD)
        print('Successfully updated your Host Identity SSH keys')
    except Exception as e:
        raise SystemExit(f'Unable to save keys to disk - Error: {e}')


def refresh_user_key(ctx, cid: str, api: AspenAPI):
    """
    Create / update User identity keys and save to default location.
    """
    Path(f'~/.ssh/').expanduser().mkdir(mode=755, exist_ok=True)
    keys = generate_openssh_keys(encrypted=True)

    data = OpenSshSigningRequestData(
        public_key=keys.public.decode('utf8'),
        cert_type=SSHCertificateType.USER,
        duration=settings.DEVICE_USER_KEY_DURATION,
    )

    try:
        resp = api.post(cid, post_data=data, subpath=settings.REFRESH_SSH_KEY_PATH)
    except ConsoleAPINotFoundError:
        url = api.get_url(cid, subpath=settings.REFRESH_SSH_KEY_PATH)
        raise SystemExit(f'Unable to refresh user identity key - URL not found {url!r}')
    except LCHTTPError as e:
        display_errors(e.response)
        raise SystemExit(f'Unable to refresh user identity key - Error: {e}')

    signed_key = resp.json()['signed_key'].encode('utf8')

    try:
        persist_user_keys(ctx, keys, signed_key=signed_key)
        confirm_known_hosts_contains_cas(api)
        print('Successfully updated your User Identity SSH keys')
    except Exception as e:
        raise SystemExit(f'Unable to save keys to disk - Error: {e}')


def generate_openssh_keys(encrypted: bool = True) -> Key:
    """
    Generate Private and Public OpenSSH keys.

    :param encrypted: Whether or not to encrypt the private key.
        Host private keys will not be encrypted.
    """
    if encrypted:
        token_password = secrets.token_urlsafe().encode('utf8')
        algorithm = BestAvailableEncryption(token_password)
    else:
        token_password = b''
        algorithm = NoEncryption()

    ed_private_key = Ed25519PrivateKey.generate()
    private_bytes = ed_private_key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.OpenSSH,
        encryption_algorithm=algorithm,
    )

    public_bytes = ed_private_key.public_key().public_bytes(
        encoding=Encoding.OpenSSH,
        format=PublicFormat.OpenSSH,
    )

    return Key(private_bytes, public_bytes, token_password.decode('utf8'))


def persist_user_keys(ctx, keys: Key, signed_key: bytes):
    """
    Save new keys identity keys to standard ssh location,
    but with custom key names, so that they'll need to be
    added manually by aspen calls, with aspen-managed password
    for private key encryption.
    """
    private_key_path = Path(f'~/.ssh/{settings.USER_KEY_NAME}').expanduser()

    flush_private_key_identity(ctx)

    ssh_key_files = (
        (private_key_path, keys.private),
        (f'{private_key_path}.pub', keys.public),
        (f'{private_key_path}-cert.pub', signed_key),
    )

    for key_file, content in ssh_key_files:
        try:
            os.chmod(key_file, stat.S_IWRITE)
        except FileNotFoundError:
            pass

        with open(key_file, 'wb') as pk:
            pk.write(content)

        os.chmod(key_file, stat.S_IREAD)

    save_secret(settings.ASPEN_USER_PK_PWD, keys.password)
    add_private_key_identity(ctx, private_key_path, keys.password)


def persist_host_keys(ctx, key: Key, signed_key: bytes):
    """
    Save new host identity keys configuration files that
    will be referenced in /etc/ssh/sshd_config as:

        HostKey /etc/ssh/ssh_compass_host
        HostCertificate /etc/ssh/ssh_compass_host.pub

    For the keys above to work, the following files are populated as well,
    based on whether this is a Device or Appliance / Server.
        AuthorizedPrincipalsFile /etc/ssh/authorized_principals/%u
        TrustedUserCAKeys /etc/ssh/compass_user_ca.pub
    """
    if settings.IS_DRAWBRIDGE_OR_SERVER:
        _persist_server_host_keys(key, signed_key)
    else:
        _persist_device_host_keys(ctx, key, signed_key)

    update_authorized_principals(ctx)
    confirm_sshd_config_file(ctx)
    maintain_user_ca_public_keys(ctx)


def update_authorized_principals(ctx):
    """
    Authorized Principals are required for the Authorization side of the
    Authentication / Authorization coin.

    The following files are populated based on whether this is a Device or Appliance / Server.
        AuthorizedPrincipalsFile /etc/ssh/authorized_principals/%u
        TrustedUserCAKeys /etc/ssh/compass_user_ca.pub
    """
    if settings.IS_DRAWBRIDGE_OR_SERVER:
        add_server_authorized_principals()
    else:
        add_device_authorized_principals(ctx)


def _persist_server_host_keys(key: Key, signed_key: bytes):
    """
    Save Appliance / Server's host identity keys.

    The AuthorizedPrincipalsFile is ensured to contain the principals extracted
    from the network / nginx config files, so that anyone with those principals
    can access the server.
    """
    ssh_key_files = (
        (f'{settings.SSH_DIR}/{settings.HOST_KEY_NAME}', key.private),
        (f'{settings.SSH_DIR}/{settings.HOST_KEY_NAME}.pub', signed_key),
    )

    for file, content in ssh_key_files:
        try:
            os.chmod(file, stat.S_IWRITE)
        except FileNotFoundError:
            pass

        with open(file, 'wb') as pk:
            pk.write(content)

        os.chmod(file, stat.S_IREAD)


def _persist_device_host_keys(ctx, key: Key, signed_key: bytes):
    """
    Save device's host identity keys, calling sudo to mv files to /etc/ssh.

    The AuthorizedPrincipalsFile is ensured to contain the Device Owner's Canonical ID
    """
    ssh_key_files = (
        (Path(f'~/{settings.HOST_KEY_NAME}').expanduser(), key.private),
        (Path(f'~/{settings.HOST_KEY_NAME}.pub').expanduser(), signed_key),
    )
    sudo = 'sudo' if getpass.getuser() != 'root' else ''

    for file, content in ssh_key_files:
        with open(file, 'wb') as pk:
            pk.write(content)

        ctx.run(f'{sudo} rm -f {settings.SSH_DIR}/{file.name}')
        ctx.run(f'{sudo} mv {file} {settings.SSH_DIR}/{file.name}')
        ctx.run(f'{sudo} chmod 400 {settings.SSH_DIR}/{file.name}')


def add_device_authorized_principals(ctx):
    """
    Personal Devices should not be accessible by anyone on the team,
    so create an AuthorizedPrincipalsFile containing their Canonical ID.
    """
    authorized = f'{settings.SSH_DIR}/authorized_principals'
    sudo = 'sudo' if getpass.getuser() != 'root' else ''

    ctx.run(f'{sudo} mkdir -p {authorized}')
    current_user = getpass.getuser()

    try:
        with open(f'{authorized}/{current_user}', 'r') as af:
            authorized_principals = [p.strip() for p in af.readlines()]
    except FileNotFoundError:
        authorized_principals = []

    cfg = AspenCtlSettingsConfig()
    cfg_data = cfg.as_typed_dict(settings.ACCOUNT_SECTION)
    owner_cid = cfg_data.get('cid')
    if owner_cid and owner_cid not in authorized_principals:
        authorized_principals.append(owner_cid)

    if not authorized_principals:
        return

    # write to temp file, because we can't sudo redirect as that raises permission error
    tmp = f'{tempfile.gettempdir()}/authorized_principals'
    with open(tmp, 'w') as tpf:
        for principal in authorized_principals:
            tpf.write(f'{principal}\n')

    ctx.run(f"{sudo} mv {tmp} {authorized}/{current_user}")


def add_server_authorized_principals():
    """
    Add network-extracted authorized principals
    """
    os.makedirs(f'{settings.SSH_DIR}/authorized_principals', exist_ok=True)

    # DrawBridgeOS systems never login as root!
    current_user = 'dbos' if settings.IS_DRAWBRIDGE_OS else getpass.getuser()

    try:
        with open(f'{settings.SSH_DIR}/authorized_principals/{current_user}', 'r') as af:
            authorized_principals = set(p.strip() for p in af.readlines())
    except FileNotFoundError:
        authorized_principals = set()

    authorized_principals.update(get_network_ssh_principals())
    authorized_principals.add('Sysadmin')

    if settings.IS_DRAWBRIDGE_FILTER:
        authorized_principals.add('DrawBridge')

    # should be running as root, so we don't need to sudo
    with open(f'{settings.SSH_DIR}/authorized_principals/{current_user}', 'w') as af:
        for principal in authorized_principals:
            af.write(f'{principal}\n')


def maintain_user_ca_public_keys(ctx):
    """
    Ensure that Trusted User CA file exists on disk
    and contains all available User CA Public Keys.

        TrustedUserCAKeys /etc/ssh/compass_user_ca.pub
    """
    api, _ = get_api(ctx)
    api.load('aspentls', 'certificateauthority')

    user_ca_public_keys = set(retrieve_ssh_ca_public_keys(api, settings.USER_CA_PUBLIC_KEY_SUBPATH))

    try:
        with open(settings.USER_CA_PUB_FILE) as caf:
            current_cas = set(ca.strip() for ca in caf.readlines())
    except FileNotFoundError:
        current_cas = set()
    except Exception as e:
        raise SystemExit(f'Failed to read User CA Public Keys: {settings.USER_CA_PUB_FILE} - {e}')

    if user_ca_public_keys == current_cas:
        return False

    tmp = f'{tempfile.gettempdir()}/user_ca'

    try:
        with open(tmp, 'w') as tca:
            tca.writelines([f'{ca}\n' for ca in user_ca_public_keys])
    except Exception as e:
        raise SystemExit(f'Unable to save User CA Public Keys: {settings.USER_CA_PUB_FILE} - {e}')

    sudo = 'sudo' if getpass.getuser() != 'root' else ''
    ctx.run(f'{sudo} mv {tmp} {settings.USER_CA_PUB_FILE}')

    return True


def confirm_sshd_config_file(ctx):
    """
    Confirm that sshd_config file has needed lines.
    """
    needful_lines = (
        f'HostKey {settings.SSH_DIR}/{settings.HOST_KEY_NAME}',
        f'HostCertificate {settings.SSH_DIR}/{settings.HOST_KEY_NAME}.pub',
        f'AuthorizedPrincipalsFile {settings.SSH_DIR}/authorized_principals/%u',
        f'TrustedUserCAKeys {settings.USER_CA_PUB_FILE}',
    )
    with open(settings.SSH_CONFIG_FILE, 'r') as cf:
        config_lines = [line.strip() for line in cf.readlines()]

    missing = []
    for line in needful_lines:
        if line not in config_lines:
            missing.append(line)

    if missing:
        if not settings.IS_DRAWBRIDGE_OR_SERVER:
            print('\n# -----------------------------------------\n')
            print(f'Add the following lines to {settings.SSH_CONFIG_FILE}\n')
            for line in missing:
                print(line)
            print('\n# -----------------------------------------\n')
            return

        config_lines.extend(missing)
        with open(settings.SSH_CONFIG_FILE, 'w') as cf:
            for line in config_lines:
                cf.write(f'{line}\n')


def clear_sshd_config_file(ctx):
    """
    Confirm that sshd_config file has removed all Aspen CA lines.
    Needs to be called after a device / server is unregistered.
    """

    needful_lines = (
        f'HostKey {settings.SSH_DIR}/{settings.HOST_KEY_NAME}',
        f'HostCertificate {settings.SSH_DIR}/{settings.HOST_KEY_NAME}.pub',
        f'AuthorizedPrincipalsFile {settings.SSH_DIR}/authorized_principals/%u',
        f'TrustedUserCAKeys {settings.USER_CA_PUB_FILE}',
    )
    with open(settings.SSH_CONFIG_FILE, 'r') as cf:
        config_lines = [line.strip() for line in cf.readlines()]

    yanked_lines = []
    for line in needful_lines:
        if line in config_lines.copy():
            yanked_lines.append(line)
            config_lines.remove(line)

    if yanked_lines:
        if not settings.IS_DRAWBRIDGE_OR_SERVER:
            print('\n# -----------------------------------------\n')
            print(f'Remove the following lines from {settings.SSH_CONFIG_FILE}\n')
            for line in yanked_lines:
                print(line)
            print('\n# -----------------------------------------\n')
            return

        with open(settings.SSH_CONFIG_FILE, 'w') as cf:
            for line in config_lines:
                cf.write(f'{line}\n')

        ctx.run(settings.SSH_SERVICE_CMD)


def confirm_known_hosts_contains_cas(api):
    """
    Confirm that Aspen Host Certificate Authorities are in the
    user's known_hosts file, so that devices are correctly
    trusted on first use.
    """
    known_hosts_path = Path('~/.ssh/known_hosts').expanduser()
    api.load('aspentls', 'certificateauthority')

    host_ca_public_keys = set(
        f'@cert-authority * {key}'.strip()
        for key in retrieve_ssh_ca_public_keys(api, settings.HOST_CA_PUBLIC_KEY_SUBPATH)
    )

    try:
        with open(known_hosts_path, 'r') as kn:
            known_hosts = [h.strip() for h in kn.readlines()]
    except FileNotFoundError:
        known_hosts = set()

    current_cas = set()
    host_lines = []

    for entry in known_hosts:
        if entry.startswith('@cert-authority *'):
            current_cas.add(entry)
        else:
            host_lines.append(entry)

    if current_cas == host_ca_public_keys:
        return

    with open(known_hosts_path, 'w') as kn:
        kn.writelines([f'{host}\n' for host in host_ca_public_keys])
        kn.writelines([f'{host}\n' for host in host_lines])


def retrieve_ssh_ca_public_keys(api, subpath) -> Set[str]:
    """
    Helper method to retrieve Host / User public keys from Aspen server.
    """
    api.load('aspentls', 'certificateauthority')

    try:
        resp = api.get(subpath=subpath)
        return resp.json()['public_keys']
    except KeyError:
        raise SystemExit(f'No Host CA Public Keys found in {resp.json()}')
    except LCHTTPError as e:
        display_errors(e.response)
        raise SystemExit(f'Unable to retrieve Host CA Public Keys - Error: {e}')


def add_private_key_identity(
    ctx,
    path: Path,
    password: str = None,
):
    """
    Add the private key to available SSH identities.
    """
    if not path.exists():
        print(f'{path} not found')
        return

    secret = password or load_secret(settings.ASPEN_USER_PK_PWD)
    responder = Responder(pattern=r"Enter passphrase", response=f'{secret}\n')

    add_cmd = 'keychain -q --nogui' if settings.CYGWIN_OR_WSL else 'ssh-add -q'

    try:
        ctx.run(f'{add_cmd} {path}', watchers=[responder], pty=True)
    except Exception:  # noqa
        print(
            f'Unable to add {path} to SSH agent. '
            f'Ensure "ssh-add" or "keychain" command is available'
        )


def flush_private_key_identity(ctx):
    """
    Flush current private key from SSH identities, so that when a refreshed
    private / public key pair has been generated, the current values will be
    destroyed. Otherwise, the ssh-agent selects the previous identity.
    """
    key_file = Path(f'~/.ssh/{settings.USER_KEY_NAME}').expanduser()
    if not key_file.exists():
        return

    flush_cmd = 'keychain -q --clear' if settings.CYGWIN_OR_WSL else f'ssh-add -q -d {key_file}'

    try:
        ctx.run(flush_cmd, pty=True)
    except Exception:  # noqa
        pass  # key may not have been added so go on silently


def purge_ssh_identity_keys(ctx):
    """
    Remove SSH identity keys and flush the ssh-agent.
    """
    flush_private_key_identity(ctx)

    for key_file in (
            f'~/.ssh/{settings.USER_KEY_NAME}',
            f'~/.ssh/{settings.USER_KEY_NAME}.pub',
            f'~/.ssh/{settings.USER_KEY_NAME}-cert.pub',
            f'{settings.SSH_DIR}/{settings.HOST_KEY_NAME}',
            f'{settings.SSH_DIR}/{settings.HOST_KEY_NAME}.pub',
            settings.USER_CA_PUB_FILE,
    ):
        ctx.sudo(f'rm -f {key_file}')


def expired_ssh_certificate(
    path: Union[str, Path],
    renewal_margin: int = 0,
) -> Tuple[bool, Optional[SSHCertificate]]:
    """
    Check if SSH certificate "valid before" date is in the past.

    :param path: Path location to SSH certificate file.
    :param renewal_margin: Number of days before the real expiration
    to consider the cert "expired". Host certificates should be
    renewed some days before the actual expiration, to ensure
    the host remains reachable.
    """
    try:
        with open(path, 'rb') as f:
            cert = SSHCertificate.from_bytes(f.read())
    except (FileNotFoundError, SSHCertificateParserError):
        return True, None

    if renewal_margin > 0:
        renewal_margin = int((datetime.utcnow() - timedelta(days=renewal_margin)).timestamp())

    return cert.remaining_validity <= renewal_margin, cert


__all__ = (
    'refresh_host_key',
    'refresh_user_key',
    'add_private_key_identity',
    'maintain_user_ca_public_keys',
    'purge_ssh_identity_keys',
    'clear_sshd_config_file',
    'expired_ssh_certificate',
    'update_authorized_principals',
)
