from lcconfig import AspenCtlSettingsConfig, strtobool
from string import punctuation
from uuid import UUID

from ...network import get_api
from ... import settings

table = str.maketrans(punctuation, ' ' * len(punctuation))


def register(ctx):
    """
    Register this device on Aspen CA to get signed Public Keys.
    """
    cfg = AspenCtlSettingsConfig()
    data = cfg.as_typed_dict(settings.DEVICE_SECTION)

    if data.get('cid'):
        print(f'Device {data["name"]!r} already exists')
        return

    api, user = get_api(ctx)
    api.load('devices', 'device')

    link_to_existing = None
    while link_to_existing is None:
        try:
            link_to_existing = strtobool(input('Link to existing device? (y/n): '))
        except (EOFError, KeyboardInterrupt):
            print()
            raise SystemExit()
        except (AttributeError, ValueError):
            continue

    if link_to_existing:
        device_cid = None

        while not device_cid:
            try:
                device_cid = str(UUID(input('Device Cid: ')))
            except (AttributeError, ValueError, TypeError):
                print('Not a valid Canonical ID')
                continue

            try:
                resp = api.get(device_cid)
                resp_data = resp.json()
                device_data = {
                    'cid': device_cid,
                    'name': resp_data['name'],
                }

                cfg.save_section(settings.DEVICE_SECTION, device_data)
                return
            except (EOFError, KeyboardInterrupt):
                print()
                raise SystemExit()
            except Exception as e:
                raise SystemExit(
                    f'Unable to confirm that device with CID {device_cid} exists - Error: {e}'
                )

    data = {
        'contact': user,
        'company': settings.COMPASS_COMPANY_CID,
    }

    device_name = None
    while not device_name:

        try:
            name = input('Specify a device name: ')
            if name and name[0].isdigit():
                print('Name cannot start with a number')
                continue
            if len(name) < 7:
                print('Name must be at least 7 characters long')
                continue
        except (EOFError, KeyboardInterrupt):
            print()
            raise SystemExit()

        device_name = name.translate(table)
        data['name'] = device_name

    try:
        resp = api.post(post_data=data)
        if resp.ok:
            print(f'Created device {resp.json()["name"]!r}')
        else:
            print('Unable to create device')
            return

        resp_data = resp.json()
        device_data = {
            'cid': resp_data['cid'],
            'name': resp_data['name'],
        }

        cfg.save_section(settings.DEVICE_SECTION, device_data)

    except Exception as e:
        raise SystemExit(f'An error occurred when creating this device - Error: {e}')


__all__ = [
    'register',
]
