from invoke import task
from lcconfig import strtobool
import os
from pathlib import Path
from uuid import uuid1, UUID

from aspenctl.configuration import ConsoleSettingsConfig
from aspenctl.network import get_api
from aspenctl.ssh.principals import get_network_ssh_principals
from aspenctl import settings
from aspenctl.typehints import ApplianceData
from aspenctl.rsa import set_ca_defaults
from aspenctl.validation import is_alphabetic, is_valid_hostname


def register(ctx):
    """
    Create new server on Aspen, or link to existing server.
    """
    ctx[settings.PERSIST_TOKEN] = False

    cfg = ConsoleSettingsConfig(ctx=ctx)
    cfg_data = cfg.as_typed_dict(settings.SYSTEM_SECTION)
    system_cid = cfg_data.get('cid')

    if system_cid:
        print('System has already been registered')
        return cfg_data

    link_to_existing = None
    while link_to_existing is None:
        try:
            link_to_existing = strtobool(input('Link to existing device? (y/n): '))
        except (EOFError, KeyboardInterrupt):
            print()
            raise SystemExit()
        except (AttributeError, ValueError):
            continue

    set_ca_defaults()

    if not link_to_existing:
        return create_new_aspen_record(ctx)

    return relink_to_aspen_record(ctx)


def relink_to_aspen_record(ctx):
    """
    Link this system to an existing record in Aspen.
    """
    appliance_cid = False
    while not appliance_cid:
        try:
            appliance_cid = str(UUID(input('Appliance / Server Cid: ')))
        except (EOFError, KeyboardInterrupt):
            print()
            raise SystemExit()
        except (AttributeError, ValueError, TypeError):
            print('Not a valid Canonical ID')
            continue

    api, _ = get_api(ctx)
    api.load('accounts', 'appliance')

    try:
        resp = api.get(appliance_cid)
        resp_data = resp.json()

        if ctx.get(settings.PERSIST_TOKEN, True):
            try:
                auth_token = api.get(appliance_cid, subpath='get_token').json().get('auth_token')
            except Exception as e:
                auth_token = ''
                print(f'Unable to retrieve auth token - Error: {e}')
        else:
            auth_token = ''

        data = ApplianceData(
            adminurl=resp_data['adminurl'],
            auth_token=auth_token,
            name=resp_data['name'],
            cid=appliance_cid,
        )

        cfg = ConsoleSettingsConfig(ctx=ctx)
        cfg.save_section(settings.SYSTEM_SECTION, data)
        return data
    except Exception as e:
        print(f'Unable to confirm that Appliance / Server with CID {appliance_cid} exists\n\n{e}')
        return ApplianceData()  # noqa


def create_new_aspen_record(ctx):
    """
    Generate, register and return settings.conf data.
    """
    print('Registering this server on Aspen')

    name = ''
    adminurl = ''

    while not name:
        _name = input('Enter system name: ')
        if is_alphabetic(_name):
            name = _name
        else:
            print(f'Contains invalid characters: {_name!r}. Please try again.')

    while not adminurl:
        hostname = input('Admin hostname: ')
        if is_valid_hostname(hostname):
            adminurl = hostname
        else:
            print(f'Invalid hostname: {hostname!r}. Please try again.')

    data = {
        'name': name,
        'adminurl': adminurl,
        'cid': str(uuid1()),
        'company': settings.COMPASS_COMPANY_CID,
    }

    os.makedirs(settings.COMPASS_CONFIG_DIR, exist_ok=True)
    cfg = ConsoleSettingsConfig(ctx=ctx)
    cfg.save_section(settings.SYSTEM_SECTION, ApplianceData(**data))

    api, _ = get_api(ctx, account_cid=data['cid'])
    api.load('accounts', 'appliance')

    try:
        api.put(data['cid'], post_data=data)
        if ctx.get(settings.PERSIST_TOKEN, True):
            try:
                data['auth_token'] = api.get(data['cid'], subpath='get_token')
            except Exception as e:
                data['auth_token'] = ''
                print(f'Unable to retrieve auth token - Error: {e}')
        data.pop('company')
    except Exception as e:
        try:
            Path(cfg.persist_file).unlink()
        except FileNotFoundError:
            pass
        raise SystemExit(f'Unable to register Appliance / Server - Error: {e}')

    add_principals(ctx)

    return data


@task
def add_principals(ctx):  # noqa
    """
    Add more principals to host's identity key
    """
    ssh_principals = get_network_ssh_principals()
    print('Current host principals are: ')
    for principal in ssh_principals:
        print(principal)

    new_principals = []
    update_principals = True

    while update_principals:
        try:
            update_principals = strtobool(input('Add additional principal: (y/n) '))
        except (EOFError, KeyboardInterrupt):
            print()
            raise SystemExit()
        except (AttributeError, ValueError):
            continue

        if update_principals:
            principal = input('Enter new SSH Principal: ').strip()
            if principal in ssh_principals:
                print(f'Principal already exists: {principal!r}')
                continue

            new_principals.append(principal)

    if new_principals:
        ssh_principals.extend(new_principals)
        cfg = ConsoleSettingsConfig()
        cfg_data = cfg.as_typed_dict(settings.SYSTEM_SECTION)
        cfg_data['principals'] = ','.join(ssh_principals)

        cfg.save_section(settings.SYSTEM_SECTION, cfg_data)

    return ssh_principals


__all__ = (
    'register',
    'relink_to_aspen_record',
    'create_new_aspen_record',
    'add_principals',
)
