from aspenctl.configuration import AspenCtlSettingsConfig, ConsoleSettingsConfig
from invoke import task
from aspenctl.network import get_api
from aspenctl.ssh import refresh_host_key, expired_ssh_certificate, maintain_user_ca_public_keys
from aspenctl import settings


@task(
    name='host',
    help={
        'force': 'Renew whether expired or not',
        'cron': 'Called from cron job or other task scheduler',
    },
)
def host_identity_key(ctx, force=False, cron=False):
    """
    Get or renew a signed Host Identity key for this Device or Server.
    """
    # Returning a bool value to indicate whether a refresh occurred or not so that
    # when invoked from the Log Cabin console we know whether to restart OpenSSH or not.

    if not force:
        cert_path = f'{settings.SSH_DIR}/{settings.HOST_KEY_NAME}.pub'
        expired, cert = expired_ssh_certificate(cert_path, settings.HOST_RENEWAL_MARGIN_DAYS)
        if not expired:
            # Always check for User CA changes to handle additions and revocations
            if maintain_user_ca_public_keys(ctx):
                ctx.run(settings.SSH_SERVICE_CMD)
            if not cron:
                print(
                    f'Skipping renewal - current certificate is valid till {cert.valid_before}.\n'
                    f'To renew anyway, pass in the --force flag.'
                )
            return False

    api, _ = get_api(ctx)

    if settings.IS_DRAWBRIDGE_OR_SERVER:
        cid = server_canonical_id(ctx)
        api.load('accounts', 'appliance')
    else:
        cid = device_canonical_id(ctx)
        api.load('devices', 'device')

    refresh_host_key(ctx, cid, api)

    return True


def device_canonical_id(ctx):
    """
    Get Canonical ID for this Device.
    """
    print('Get SSH host identity key for this device')

    cfg = AspenCtlSettingsConfig()
    cfg_data = cfg.as_typed_dict(settings.DEVICE_SECTION)

    try:
        return cfg_data['cid']
    except KeyError:
        pass

    try:
        from ..register.device import register
        return register(ctx)['cid']
    except KeyError:
        raise SystemExit('Unable to register this Device on Aspen')


def server_canonical_id(ctx):
    """
    Get Canonical ID for this Appliance / Server.
    """
    print('Get SSH host identity key for this Appliance / Server')

    cfg = ConsoleSettingsConfig()
    cfg_data = cfg.as_typed_dict(settings.SYSTEM_SECTION)

    try:
        return cfg_data['cid']
    except KeyError:
        pass

    try:
        from ..register.server import register
        return register(ctx)['cid']
    except KeyError:
        raise SystemExit('Unable to register this Appliance / Server on Aspen')


__all__ = [
    'host_identity_key',
]
