from invoke import task
from lcconfig import ConsoleSettingsConfig

from aspenctl.authorities import ca_selector
from aspenctl.network import get_api
from aspenctl.rsa import (
    generate_csr,
    get_common_name,
    get_alt_dns_names,
    get_ip_addresses,
    get_email,
    get_value_from_user,
    sign_csr,
    save_csr_conf,
    csr_history,
    renewal_required,
)
from aspenctl.settings import ASPEN_CONFIG_DIR


@task(
    help={'location': f'Location on disk to save signed certificate. Default: {ASPEN_CONFIG_DIR}'},
)
def csr(ctx, location=ASPEN_CONFIG_DIR):
    """
    Create a Certificate Signing Request and get it signed by Aspen
    """
    api, _ = get_api(ctx)
    api.load('aspentls', 'rootcertificateauthority')

    common_name = get_common_name()
    ca_pk, ca_name = ca_selector(api, 'Select CA to sign this request')

    cfg = ConsoleSettingsConfig()
    data = cfg.as_typed_dict('CA')

    generate_csr(
        common_name=common_name,
        country=get_value_from_user('Country', 'US'),
        state=get_value_from_user('State', data.get('state')),
        locality=get_value_from_user('Locality', data.get('locality')),
        organization=get_value_from_user('Organization / Company', data.get('organization')),
        email=get_email(data.get('email')),
        dnsnames=get_alt_dns_names(),
        ips=get_ip_addresses(),
    )

    expires = int(get_value_from_user('Days till expiration', 820))
    save_csr_conf(common_name, expires, ca_pk, location)
    sign_csr(ctx, common_name, expires, ca_pk)

    print(f'\nGenerated CSR and Signed Certificate for {common_name}')

    for ext in ('.pem', '.key', '.csr'):
        print(f'{location}/{common_name}{ext}')


@task(name='sign', help={'common_name': 'Common name of the CSR to be signed / renewed'})
def sign_request(ctx, common_name):
    """
    Submit a Certificate Signing Request to Aspen for initial signing or renewal
    """
    for ch in csr_history():
        if ch['name'] == f'{common_name}.csr':
            print(f"Renewing {common_name}")
            sign_csr(ctx, common_name, ch['expires'], ch['ca'])


@task(help={'cron': 'Set to True if called from cron, to reduce print output'})
def renew_all(ctx, cron=False):
    """
    Renew all Certificates that have a valid Certificate Signing Request.
    Can be called via cron job to keep all certificates up to date.
    """
    for ch in csr_history():
        name = ch['name']

        if not renewal_required(f'{ch["location"]}/{name}.pem'):
            if not cron:
                print(f'Cert for {name} not yet expired')
            continue

        if not cron:
            print(f"Renewing {name}")

        sign_csr(ctx, name, ch['expires'], ch['ca'])


__all__ = (
    'csr',
    'sign_request',
    'renew_all',
)
