from ipaddress import IPv4Address
import re


def is_alphabetic(value: str) -> bool:
    """
    Confirm that specified value has no special characters.
    """
    return value.isalnum()


def is_valid_hostname(hostname: str) -> bool:
    """
    Confirm that specified hostname is valid. Doesn't expect schema
    """
    if hostname[-1] == ".":
        # strip exactly one dot from the right, if present
        hostname = hostname[:-1]
    if len(hostname) > 253:
        return False

    labels = hostname.split(".")

    # the TLD must be not all-numeric
    if re.match(r"[0-9]+$", labels[-1]):
        return False

    allowed = re.compile(r"(?!-)[a-z0-9-]{1,63}(?<!-)$", re.IGNORECASE)
    return all(allowed.match(label) for label in labels)


def is_valid_ip_address(value: str) -> bool:
    """
    Confirm that specified value is a valid IPv4 address.
    """

    try:
        return bool(IPv4Address(value))
    except Exception:
        return False


def is_valid_email_address(value: str) -> bool:
    """
    Confirm that specified value is a valid Email address.
    """

    try:
        return bool(re.match(r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)", value, re.I))
    except Exception:
        return False


def display_errors(response):
    """
    Print out errors to console.
    """
    try:
        errors = response.json()
        print()
        for field, messages in errors.items():
            print(f'{field} - {messages}')
        print()
    except AttributeError:
        pass


__all__ = (
    'is_alphabetic',
    'is_valid_hostname',
    'is_valid_ip_address',
    'is_valid_email_address',
    'display_errors',
)
