from cryptography.hazmat.primitives import serialization
from cryptography import x509
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives import hashes

from invoke import run
import os
from pathlib import Path
from pylogcabin.exceptions import ConsoleAPIAuthenticationError, ConsoleAPIError

from ..network import get_api
from ..rsa.keys import generate_key
from aspenctl import settings


# ---------------------------------------------------------------------------
def generate_appliance_vpn_csr(cid: str) -> str:
    """
    Generate a DrawBridge Appliance's VPN certificate signing request and private key.
    Returns a path to the saved csr certificate.
    """
    os.makedirs(settings.VPN_KEYS_DIR, exist_ok=True)
    private_key = generate_key(name='management_vpn', path=settings.VPN_KEYS_DIR)

    locality_name = 'DrawBridge' if settings.IS_DRAWBRIDGE_FILTER else 'Compass'

    subject_name = x509.Name([
        # Include all the details to be shown in the cert
        x509.NameAttribute(NameOID.COUNTRY_NAME, 'US'),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, 'PA'),
        x509.NameAttribute(NameOID.LOCALITY_NAME, f'{locality_name} Systems'),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, 'Compass Foundation, LLC'),
        x509.NameAttribute(NameOID.COMMON_NAME, cid),
    ])

    csr = x509.CertificateSigningRequestBuilder(
        subject_name=subject_name,
    ).add_extension(
        x509.KeyUsage(
            digital_signature=True,
            content_commitment=False,
            key_encipherment=True,
            data_encipherment=True,
            key_agreement=True,
            key_cert_sign=False,
            crl_sign=False,
            encipher_only=False,
            decipher_only=False,
        ),
        critical=True,
    ).add_extension(
        x509.BasicConstraints(ca=False, path_length=None), critical=True
    ).add_extension(
        x509.ExtendedKeyUsage([x509.OID_CLIENT_AUTH]), critical=True
    ).sign(private_key, hashes.SHA512())

    csr_file = f'{settings.VPN_KEYS_DIR}/vpn.csr'
    with open(csr_file, 'wb') as f:
        f.write(csr.public_bytes(serialization.Encoding.PEM))

    return csr_file


# ---------------------------------------------------------------------------
def sign_appliance_vpn_csr(ctx, cid, csr_path: str):
    """
    Sign a DrawBridge Appliance's VPN certificate signing request and save certificate.

    Expecting a dictionary in response, with signed cert and VPN CA and static keys.
    {
        'pem': data,
        'ca': ca.pub.pem,
        'key': vpn_static_key(),
    }
    """
    try:
        with open(csr_path, 'r') as csrp:
            csr = csrp.read()
    except Exception as e:
        raise SystemExit(e)

    api, user = get_api(ctx)
    api.load('accounts', 'appliance')

    try:
        resp = api.post(cid, subpath='sign_vpn_csr', post_data={'pem': csr})
    except ConsoleAPIAuthenticationError:
        raise SystemExit(f'Invalid credentials; unable to sign VPN CSR.')
    except ConsoleAPIError as e:
        raise SystemExit(f'Error signing request {e.response.text}')
    except Exception as e:
        raise SystemExit(f'Failed to sign CSR on Aspen CA Server: {e}')

    try:
        _save_vpn_certs(resp.json())
    except Exception as e:
        raise SystemExit(f'Unable to save VPN certs to disk: {e}')

    Path(f'{settings.ASPEN_CONFIG_DIR}/vpn.csr').unlink(missing_ok=True)


# ---------------------------------------------------------------------------
def _save_vpn_certs(data):
    """
    Expecting a dictionary in response to certificate signing request to
    Aspen, with signed cert and VPN CA and static keys.
    {
        'pem': '-----BEGIN CERTIFICATE-----\nMIIFqzCCA...SCaOi3F1ty\n-----END CERTIFICATE-----',
        'ca': '-----BEGIN CERTIFICATE-----\nMIIGbjCCBF...bgGH5h8AsB\n-----END CERTIFICATE-----',
        'key': '-----BEGIN OpenVPN Static key V1----- ... -----END OpenVPN Static key V1-----',
    }
    """
    name = 'drawbridge' if settings.IS_DRAWBRIDGE_FILTER else 'management_vpn'

    os.makedirs(settings.VPN_KEYS_DIR, exist_ok=True)
    for k, name in (
        ('pem', f'{settings.VPN_KEYS_DIR}/{name}.pem'),
        ('ca', f'{settings.VPN_KEYS_DIR}/{name}_ca.pem'),
        ('key', f'{settings.VPN_KEYS_DIR}/{name}_ta.key'),
        ('config', f'{settings.VPN_DIR}/{name}-client.conf'),
    ):
        with open(name, 'w') as vcf:
            vcf.write(data[k])

    run(f'chown -R openvpn:openvpn {settings.VPN_KEYS_DIR}')
    run(f'chmod 640 {settings.VPN_KEYS_DIR}/*')


__all__ = (
    'generate_appliance_vpn_csr',
    'sign_appliance_vpn_csr',
)
