#!/usr/bin/python3

import io
import re
import os
from os.path import dirname, join

from setuptools import setup


def read(*names, **kwargs):
    path = join(dirname(__file__), *names)
    return io.open(path, encoding=kwargs.get('encoding', 'utf8')).read()


def get_packages(package):
    """
    Return root package and all sub-packages.
    """
    return [
        dirpath for dirpath, dirnames, filenames in os.walk(package)
        if os.path.exists(os.path.join(dirpath, '__init__.py'))
    ]


def get_package_data(package):
    """
    Return all files under the root package, that are not in a
    package themselves.
    """
    walk = [(dirpath.replace(package + os.sep, '', 1), filenames)
            for dirpath, dirnames, filenames in os.walk(package)
            if not os.path.exists(os.path.join(dirpath, '__init__.py'))]

    filepaths = []
    for base, filenames in walk:
        filepaths.extend([os.path.join(base, filename) for filename in filenames])
    return {package: filepaths}


def get_version():
    with open('aspenctl/__init__.py', 'r') as f:
        line = f.readline()
        match = re.match(r'__version__ = \'([\d\.]+)\'', line)

        if not match:
            raise ImportError("Can't read the version of Aspen Ctl")

        version = match.group(1)
        return version


setup(
    name='aspenctl',
    version=get_version(),
    license='SAP',
    description='Aspen CA Client',
    long_description='%s\n%s' % (
        re.compile('^.. start-badges.*^.. end-badges', re.M | re.S).sub('', read('README.md')),
        re.sub(':[a-z]+:`~?(.*?)`', r'``\1``', read('CHANGELOG.md'))
    ),
    long_description_content_type='text/markdown',
    author='Dave Burkholder',
    author_email='dave@compassfoundation.io',
    url='https://code.compassfoundation.io/aspen/aspenctl',
    packages=get_packages('aspenctl'),
    package_data=get_package_data('aspenctl'),
    zip_safe=True,
    options={
        'bdist_rpm': {
            'post_install': 'post_install.sh',
            'pre_install': 'pre_install.sh',
        },
    },
    classifiers=[
        'Intended Audience :: Developers',
        'License :: SAP',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
    ],
    keywords=[
        'Aspen CA',
        'DrawBridge'
        'Sysadmin',
    ],
    install_requires=[
        'aspen_ssh',
        'bcrypt',
        'cffi',
        'cryptography',
        'ifaddr',
        'invoke',
        'keyring>=23.0.1',
        'secretstorage>=3.3.1',
        'lcconfig',
        'pyjwt',
        'pylogcabin',
        'typing_extensions ; python_version <= "3.7"',
    ],
    entry_points={
        'console_scripts': [
            'aspen = aspenctl.main:program.run',
        ],
    },
)
